import contextlib
import copy
import inspect
import itertools
import json
import math
import os
import warnings
from datetime import datetime
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy
from matplotlib import cm
from matplotlib.axes._axes import Axes
from matplotlib.figure import Figure
from numpy import float16, float64, ndarray, str_
from numpy.random.mtrand import RandomState
from scipy.spatial.distance import cdist
from scipy.stats import beta
from sklearn.exceptions import ConvergenceWarning
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF
from sklearn.model_selection import cross_val_score
from sklearn.utils._testing import ignore_warnings
from sklearn.utils.extmath import cartesian
from tqdm import tqdm


class GECBase:
    def _gec_init(
        self,
        frozen: bool,
        non_optimized_init_args: List[str],
        optimization_candidate_init_args: List[str],
        categorical_hyperparameters: List[Tuple[str, List[str]]],
    ) -> None:

        self.fix_boosting_type_ = False
        self.fix_bootstrap_type_ = False

        self.frozen = frozen

        self._init_args = {
            arg: self.retrieve_hyperparameter(arg) for arg in non_optimized_init_args
        }
        self._optimization_candidate_init_args = optimization_candidate_init_args

        self.gec_hyperparameters_ = {
            "l": 1.0,
            "hyperparams_acquisition_percentile": 0.7,
            "bandit_greediness": 1.0,
            "score_evaluation_method": None,
            "maximize_score": True,
            "n_random_exploration": 5,
            "n_sample": 1000,
            "n_sample_initial": 1000,
            "best_share": 0.2,
            "distance_metric": "cityblock",
            "hyperparameters": optimization_candidate_init_args,
            "randomize": True,
        }

        self.categorical_hyperparameters = categorical_hyperparameters + [
            ("gec_bagging", ["gec_bagging_yes", "gec_bagging_no"])
        ]
        self._categorical_hyperparameter_names = [
            cat[0] for cat in self.categorical_hyperparameters
        ]

        self._set_gec_attributes()
        self._set_gec_fields()

    def _set_gec_attributes(self) -> None:

        prohibited_combinations = [
            "rf-gec_bagging_no",
            "Plain-Bayesian-gec_bagging_yes",
            "Plain-No-gec_bagging_yes",
        ]

        self._categorical_hyperparameter_combinations = [
            "-".join(y)
            for y in itertools.product(
                *[x[1] for x in self.categorical_hyperparameters]
            )
            if "-".join(y) not in prohibited_combinations
        ]

        if self.fix_boosting_type_:
            self._categorical_hyperparameter_combinations = [
                hp_comb
                for hp_comb in self._categorical_hyperparameter_combinations
                if hp_comb.startswith(self.retrieve_hyperparameter("boosting_type"))
            ]

        if self.fix_bootstrap_type_:
            self._categorical_hyperparameter_combinations = [
                hp_comb
                for hp_comb in self._categorical_hyperparameter_combinations
                if hp_comb.split("-")[1]
                == self.retrieve_hyperparameter("bootstrap_type")
            ]

        real_hyperparameters_all_across_classes = [
            ("learning_rate", np.exp(np.arange(-6.5, 0, 0.2))),
            ("num_leaves", np.exp(np.arange(2, 7.5, 0.4)).astype(int)),
            (
                "n_estimators",
                np.unique((np.exp(np.arange(1, 7.5, 0.5)) * 3).astype(int)),
            ),
            ("reg_alpha", np.exp(np.arange(-20, 4, 0.5))),
            ("reg_lambda", np.exp(np.arange(-20, 4, 0.5))),
            ("min_child_weight", np.exp(np.arange(-20, 0.0, 1.0))),
            ("min_child_samples", np.arange(2, 50, 2)),
            ("colsample_bytree", np.arange(0.1, 1.00, 0.05)),
            ("colsample_bylevel", np.arange(0.1, 1.00, 0.05)),
            ("subsample", [x / 100 for x in range(10, 101, 5)]),
        ]
        self._real_hyperparameters_all = [
            (n, r)
            for n, r in real_hyperparameters_all_across_classes
            if n in self._optimization_candidate_init_args
        ]

        self.fixed_params = {
            hyperparameter: self.retrieve_hyperparameter(hyperparameter)
            for hyperparameter, _ in self._real_hyperparameters_all
            if hyperparameter not in self.gec_hyperparameters_["hyperparameters"]
        }

        self._real_hyperparameters = [
            (hp_name, range_)
            for hp_name, range_ in self._real_hyperparameters_all
            if hp_name in self.gec_hyperparameters_["hyperparameters"]
        ]
        self._real_hyperparameters_linear = [
            (name, np.arange(-1.0, 1.0, 2 / len(values)).round(4).astype(np.float16))
            for name, values in self._real_hyperparameters
        ]

        self._real_hyperparameters_map = {
            name: dict(zip(linear_values, real_values))
            for ((name, linear_values), (_, real_values)) in zip(
                self._real_hyperparameters_linear, self._real_hyperparameters
            )
        }

        self._real_hyperparameters_map_reverse = {
            name: dict(zip(real_values, linear_values))
            for ((name, linear_values), (_, real_values)) in zip(
                self._real_hyperparameters_linear, self._real_hyperparameters
            )
        }

        self._validate_parameter_maps()

        self._real_hyperparameter_names, self._real_hyperparameter_ranges = zip(
            *self._real_hyperparameters_linear
        )

        self._real_hypermarameter_types = [
            np.array(s).dtype for _, s in self._real_hyperparameters
        ]

    def _set_gec_fields(self) -> None:
        self.kernel = RBF(self.gec_hyperparameters_["l"])
        self.hyperparameter_scores_ = {
            "inputs": [],
            "output": [],
            "means": [],
            "sigmas": [],
        }

        self.best_score_gec_ = None
        self.best_params_ = None
        self.gec_fit_params_ = None
        self.n_iterations = 0

        # parameters for bandit
        self.rewards_ = {
            c: {"a": 1, "b": 1} for c in self._categorical_hyperparameter_combinations
        }
        self.selected_arms_ = []

        self.best_scores_gec_ = {}
        self.best_params_gec_ = {}

    def tried_hyperparameters(self):
        assert np.array(self.hyperparameter_scores_["inputs"]).shape[0] == len(
            self.selected_arms_
        )

        hyperparamters = []
        for selected_arm, hyperparameter_inputs in zip(
            self.selected_arms_, self.hyperparameter_scores_["inputs"]
        ):
            args = self._process_arguments(
                self._build_arguments(selected_arm, hyperparameter_inputs)
            )
            hyperparamters.append(args)

        return hyperparamters

    @property
    def kernel(self):
        return self._kernel

    @kernel.setter
    def kernel(self, value):
        self._kernel = value
        self.gaussian = GaussianProcessRegressor(kernel=value, n_restarts_optimizer=9)

    @property
    def gec_iter_(self) -> int:
        return len(self.hyperparameter_scores_["output"])

    @classmethod
    def _cast_to_type(cls, value, type_):
        if type_ == np.float64:
            return float(value)
        elif type_ == np.int64:
            return int(value)
        elif value is None:
            return None
        else:
            raise Exception(f"type {type_} currently not supported")

    @classmethod
    def deserialize(
        cls, path: str, X: Optional[ndarray] = None, y: Optional[ndarray] = None
    ):
        """Deserialize a model and fit underlying LGBMClassifier if X and y are provided

        Parameters
        ----------
            path : str
                path to serialized GEC
            X : np.ndarray, optional
                Input feature matrix
            y : np.ndarray, optional
                Target class labels

        Returns
        -------
            gec : GEC
                deserialized model object
        """
        with open(path, "r") as f:
            representation = json.loads(f.read())

        gec = cls()
        gec.gec_hyperparameters_ = representation["gec_hyperparameters"]
        gec.rewards_ = representation["rewards"]
        gec.selected_arms_ = representation["selected_arms"]
        gec.hyperparameter_scores_ = (
            gec._convert_gaussian_process_data_from_deserialisation(
                representation["hyperparameter_scores"]
            )
        )
        gec.best_params_ = representation["best_params"]
        gec.best_score_gec_ = float(representation["best_score_gec"])
        gec.best_params_gec_ = representation["best_params_gec"]
        gec.best_scores_gec_ = representation["best_scores_gec"]
        gec.gec_fit_params_ = representation["gec_fit_params"]

        if X is not None and y is not None:
            gec._fit_best_params(X, y)
        else:
            warnings.warn(
                "If X and y are not provided, the GEC model is not fitted for inference"
            )
        return gec

    @classmethod
    def _convert_gaussian_process_data_from_deserialisation(
        cls, data_dict: Dict[str, List[Union[List[float], float]]]
    ) -> Dict[str, List[Union[List[float], float]]]:
        converted_dict = {k: list(v) for k, v in data_dict.items()}
        return converted_dict

    def serialize(self, path: str) -> None:
        """Serialize GEC model object

        Parameters
        ----------
            path : str
                path to serialize GEC to
        """
        representation = self._get_representation()

        with open(path, "w") as f:
            f.write(json.dumps(representation))

    @classmethod
    def _convert_gaussian_process_data_for_serialisation(
        cls,
        data_dict: Dict[
            str,
            Union[
                List[Union[List[float], float64, int, ndarray]],
                List[Union[List[Union[float, float64]], float64, int, ndarray]],
            ],
        ],
    ) -> Dict[
        str,
        Union[
            List[Union[List[Union[float, float64]], float64, int, List[float64]]],
            List[Union[List[float], float64, int, List[float64]]],
        ],
    ]:
        def process_value(key, value):
            if not isinstance(value, np.ndarray):
                return value
            elif key != "inputs":
                return list(value)
            else:
                return list(value.astype(np.float64))

        converted_dict = {
            k2: [process_value(k2, vv) for vv in v] for k2, v in data_dict.items()
        }

        return converted_dict

    def _get_representation(self) -> Dict[str, Any]:
        hyperparameter_scores = self._convert_gaussian_process_data_for_serialisation(
            self.hyperparameter_scores_
        )
        representation = {
            "gec_hyperparameters": self.gec_hyperparameters_,
            "rewards": self.rewards_,
            "selected_arms": self.selected_arms_,
            "hyperparameter_scores": hyperparameter_scores,
            "best_params": self.best_params_,
            "best_score_gec": self.best_score_gec_,
            "best_params_gec": self.best_params_gec_,
            "best_scores_gec": self.best_scores_gec_,
            "gec_iter": self.gec_iter_,
            "gec_fit_params": {
                k: v
                for k, v in self.gec_fit_params_.items()
                if k not in ["log_cout", "log_cerr"]
            },
        }
        return representation

    def _validate_parameter_maps(self) -> None:
        real_to_linear_to_real = {
            v: self._real_hyperparameters_map[hp][
                self._real_hyperparameters_map_reverse[hp][v]
            ]
            == v
            for hp, values in self._real_hyperparameters
            for v in values
        }

        assert np.all(np.array(real_to_linear_to_real.values())), {
            k: v for k, v in real_to_linear_to_real.items() if not v
        }

        linear_to_real_to_linear = {
            v: self._real_hyperparameters_map_reverse[hp][
                self._real_hyperparameters_map[hp][v]
            ]
            == v
            for hp, values in self._real_hyperparameters_linear
            for v in values
        }
        assert np.all(np.array(linear_to_real_to_linear.values())), {
            k: v for k, v in linear_to_real_to_linear.items() if not v
        }

    def set_gec_hyperparameters(
        self, gec_hyperparameters: Dict[str, Union[int, float, List[str]]]
    ) -> None:
        """Set the hyperparameters of the GEC optimisation process

        Parameters
        ----------
            gec_hyperparameters : dict[str, float]
                dictionary with keys that are in self.gec_hyperparameters_
        """
        assert np.all(
            np.array(
                [hp in self.gec_hyperparameters_ for hp in gec_hyperparameters.keys()]
            )
        )
        self.gec_hyperparameters_.update(gec_hyperparameters)
        self._set_gec_attributes()

    def freeze(self):
        self.frozen = True
        return self

    def unfreeze(self):
        self.frozen = False
        return self

    def set_params(self, **kwargs) -> None:
        if "frozen" in kwargs:
            self["frozen"] = kwargs.pop("frozen")
        super().set_params(**kwargs)

    def get_params(
        self, deep: bool = True
    ) -> Dict[str, Optional[Union[str, float, int, bool]]]:
        if hasattr(self, "best_params_") and self.best_params_ is not None:
            params = copy.deepcopy(self.best_params_)
        else:
            params = super().get_params(deep)
        params["frozen"] = self.frozen

        return params

    def _fit_inner(
        self, X: ndarray, y: ndarray, n_iter: int, fixed_hyperparameters: List[str]
    ):

        self.fix_boosting_type_ = "boosting_type" in fixed_hyperparameters
        self.fix_bootstrap_type_ = "bootstrap_type" in fixed_hyperparameters
        fixed_hyperparameters = [
            hp
            for hp in fixed_hyperparameters
            if hp not in ["boosting_type", "bootstrap_type"]
        ]

        if not self.frozen:
            filtered_hyperparameters = list(
                set(self.gec_hyperparameters_["hyperparameters"]).difference(
                    set(fixed_hyperparameters)
                )
            )
            self.set_gec_hyperparameters({"hyperparameters": filtered_hyperparameters})

            (
                self.best_params_gec_["search"],
                self.best_scores_gec_["search"],
            ) = self._optimise_hyperparameters(n_iter, X, y)

            best_params_grid = self._find_best_parameters()
            self.best_params_gec_["grid"] = self._process_arguments(
                self._replace_fixed_args(best_params_grid)
            )
            self.best_scores_gec_["grid"] = self.score_single_iteration(
                X, y, self.best_params_gec_["grid"]
            )

            best_params_grid_from_search = self._find_best_parameters_from_search(
                self.best_arm_, self.best_params_raw_
            )
            self.best_params_gec_["grid_from_search"] = self._process_arguments(
                self._replace_fixed_args(best_params_grid_from_search)
            )
            self.best_scores_gec_["grid_from_search"] = self.score_single_iteration(
                X, y, self.best_params_gec_["grid_from_search"]
            )

            for source, score in self.best_scores_gec_.items():
                if self.best_score_gec_ is None or score > self.best_score_gec_:
                    self.best_score_gec_ = score
                    self.best_params_ = self.best_params_gec_[source]

        self._fit_best_params(X, y)

        return self

    def _calculate_cv_score(
        self,
        X: ndarray,
        y: ndarray,
        params: Dict[str, Optional[Union[str, float, int, float64]]],
        class_: Any,
    ) -> float64:
        clf = class_(**params)
        try:
            with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
                evaluation_fn = self.gec_hyperparameters_["score_evaluation_method"]
                maximize_score = self.gec_hyperparameters_["maximize_score"]
                cross_val_scores = cross_val_score(
                    clf,
                    X,
                    y,
                    cv=5,
                    fit_params=self.gec_fit_params_,
                    scoring=evaluation_fn,
                )
                if maximize_score:
                    score = np.mean(cross_val_scores)
                else:
                    score = -np.mean(cross_val_score)
        except Exception as e:
            params_clean = {k: v for k, v in params.items() if v is not None}
            warnings.warn(
                f"Could not calculate cross val scores for parameters: {params_clean}, due to {e}"
            )
            score = np.nan

        return score

    def _optimise_hyperparameters(
        self,
        n_iter: int,
        X: ndarray,
        Y: ndarray,
        **kwargs,
    ) -> Tuple[Dict[str, Optional[Union[int, float, str]]], float64]:

        if self.gec_hyperparameters_["randomize"]:
            np.random.seed(int(datetime.now().timestamp() % 1 * 1e7))

        n_random_exploration = min(
            self.gec_hyperparameters_["n_random_exploration"], int(n_iter / 2.0)
        )

        for i in tqdm(list(range(n_iter))):
            if (i + self.gec_iter_) < n_random_exploration:
                (
                    selected_arm,
                    selected_combination,
                ) = self._get_random_hyperparameter_configuration()

                mean, sigma = None, None
            else:
                (
                    selected_arm,
                    selected_combination,
                    mean,
                    sigma,
                ) = self._select_parameters()

            arguments = self._build_arguments(selected_arm, selected_combination)

            arguments = self._replace_fixed_args(arguments)
            score = self.score_single_iteration(
                X, Y, self._process_arguments(arguments)
            )

            if np.isnan(score) == False:
                self._update_gec_fields(
                    score, arguments, selected_arm, selected_combination, mean, sigma
                )
            else:
                warnings.warn(f"score is nan for {arguments}")

        return (self.best_params_, self.best_score_gec_)

    def _get_random_hyperparameter_configuration(
        self,
    ) -> Tuple[str_, Tuple[int, float64]]:

        random_arm = np.random.choice(self._categorical_hyperparameter_combinations)
        random_combination = np.array(
            [
                np.random.choice(range_)
                for real_hyperparameter, range_ in self._real_hyperparameters_linear
            ]
        )

        return (
            random_arm,
            random_combination,
        )

    def _select_parameters(self) -> Tuple[str, ndarray, ndarray, ndarray]:

        sampled_reward = np.array(
            [
                beta.rvs(reward["a"], reward["b"])
                for arm, reward in self.rewards_.items()
                if arm in self._categorical_hyperparameter_combinations
            ]
        )
        selected_arm_index = sampled_reward.argmax()
        selected_arm = self._categorical_hyperparameter_combinations[selected_arm_index]

        sets = np.array(
            [
                np.random.choice(range_, self.gec_hyperparameters_["n_sample_initial"])
                for _, range_ in self._real_hyperparameters_linear
            ]
        )

        combinations = self._get_combinations_to_score(sets)

        assert len(combinations), sets

        if len(self.hyperparameter_scores_["inputs"]) > 0:
            self._fit_gaussian()

        mean, sigma = self.gaussian.predict(combinations, return_std=True)

        predicted_rewards = np.array(
            [
                scipy.stats.norm.ppf(
                    self.gec_hyperparameters_["hyperparams_acquisition_percentile"],
                    loc=m,
                    scale=s,
                )
                for m, s in zip(mean, sigma)
            ]
        )

        selected_combination = combinations[np.argmax(predicted_rewards)]

        return (selected_arm, selected_combination, mean, sigma)

    def _get_combinations_to_score(self, sets: ndarray) -> List[ndarray]:
        if len(self.hyperparameter_scores_["inputs"]):
            n_best = min(max(
                3, int(self.gec_iter_ * self.gec_hyperparameters_["best_share"])
            ), 20)
            best_interactions = np.argsort(
                np.array(self.hyperparameter_scores_["output"])
            )[-n_best:]

            best_hyperparameters = np.array(self.hyperparameter_scores_["inputs"])[
                best_interactions, :
            ]

            closest_hyperparameters = cdist(
                best_hyperparameters,
                sets.T,
                metric=self.gec_hyperparameters_["distance_metric"],
            ).argsort(1)[:, : self.gec_hyperparameters_["n_sample"]]
            selected_hyperparameter_indices = np.unique(
                closest_hyperparameters.flatten()
            )

            combinations = list(sets[:, selected_hyperparameter_indices].T)
        else:
            combinations = list(sets[:, : self.gec_hyperparameters_["n_sample"]].T)

        return combinations

    def _update_gec_fields(
        self,
        score: float64,
        arguments: Dict[str, Optional[Union[str, float, int, float64]]],
        selected_arm: Union[str, str_],
        selected_combination: ndarray,
        mean: Optional[ndarray],
        sigma: Optional[ndarray],
    ) -> None:

        self.selected_arms_.append(selected_arm)
        self.hyperparameter_scores_["inputs"].append(
            [float(f) for f in selected_combination]
        )
        self.hyperparameter_scores_["output"].append(score)

        if mean is not None:
            self.hyperparameter_scores_["means"].append(mean)
            self.hyperparameter_scores_["sigmas"].append(sigma)

        if self.best_score_gec_ is not None:
            score_delta = score - self.best_score_gec_
            weighted_score_delta = (
                score_delta * self.gec_hyperparameters_["bandit_greediness"]
            )
            if score_delta > 0:
                self.rewards_[selected_arm]["a"] = (
                    self.rewards_[selected_arm]["a"] + weighted_score_delta
                )
                self.best_score_gec_ = score
                self.best_params_ = self._process_arguments(arguments)
                self.best_params_raw_ = arguments
                self.best_arm_ = selected_arm

            else:
                self.rewards_[selected_arm]["b"] = (
                    self.rewards_[selected_arm]["b"] - weighted_score_delta
                )
        else:
            self.best_score_gec_ = score
            self.best_params_ = self._process_arguments(arguments)
            self.best_params_raw_ = arguments
            self.best_arm_ = selected_arm

    def _replace_fixed_args(
        self, params: Dict[str, Optional[Union[int, float, str]]]
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        if self.fix_boosting_type_:
            params["boosting_type"] = self.retrieve_hyperparameter("boosting_type")

        return params

    def _build_arguments(
        self,
        selected_arm: str,
        real_combination_linear: ndarray,
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        categorical_combination = selected_arm.split("-")
        best_predicted_combination_converted = [
            self._real_hyperparameters_map[name][value]
            for name, value in zip(
                self._real_hyperparameter_names,
                real_combination_linear,
            )
        ]

        hyperparameter_values = categorical_combination + [
            self._cast_to_type(c, t)
            for c, t in zip(
                list(best_predicted_combination_converted),
                self._real_hypermarameter_types,
            )
        ]

        arguments = dict(
            zip(
                list(self._categorical_hyperparameter_names)
                + list(self._real_hyperparameter_names),
                hyperparameter_values,
            )
        )

        return {**self.fixed_params, **self._init_args, **arguments}

    @ignore_warnings(category=ConvergenceWarning)
    def _fit_gaussian(self) -> None:
        output = np.array(self.hyperparameter_scores_["output"])
        output = (output - np.max(output)) + 1
        output[output < -1.0] = -1.0

        self.gaussian.fit(np.array(self.hyperparameter_scores_["inputs"]), output)

    def _get_best_arm(self) -> str:
        mean_reward = np.array(
            [
                reward["a"] / (reward["a"] + reward["b"])
                for arm, reward in self.rewards_.items()
                if arm in self._categorical_hyperparameter_combinations
            ]
        )
        best_arm = self._categorical_hyperparameter_combinations[mean_reward.argmax()]
        return best_arm

    def _find_best_parameters(
        self, step_sizes: List[int] = [16, 8, 4, 2, 1]
    ) -> Dict[str, Optional[Union[int, float, str]]]:

        best_arm = self._get_best_arm()

        self._fit_gaussian()
        sets = [
            list(range_[:: step_sizes[0]]) + [range_[-1]]
            for range_ in self._real_hyperparameter_ranges
        ]
        initial_combinations = np.array(list(itertools.product(*sets)))

        best_combination, _ = self._find_best_parameters_iter(initial_combinations)

        best_params = self._find_best_parameters_from_initial_parameters(
            best_arm, best_combination, step_sizes
        )

        return best_params

    def _find_best_parameters_from_search(
        self, best_arm: str, params: Dict[str, Optional[Union[int, float, str]]]
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        self._fit_gaussian()

        best_params_linear_values = [
            self._real_hyperparameters_map_reverse[name][params[name]]
            for name in self._real_hyperparameter_names
        ]
        best_params = self._find_best_parameters_from_initial_parameters(
            best_arm,
            best_params_linear_values,
            step_sizes=[4, 2, 1],
        )

        return best_params

    def _find_best_parameters_iter(
        self, combinations: ndarray
    ) -> Tuple[ndarray, float64]:

        mean = self.gaussian.predict(combinations)
        best_score = np.max(mean)
        best_combination = combinations[np.argmax(mean)]

        return best_combination, best_score

    def _find_best_parameters_from_initial_parameters(
        self,
        best_arm: str,
        best_combination: Union[List[float16], ndarray],
        step_sizes: List[int],
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        for step_size, previous_step_size in zip(step_sizes[1:], step_sizes[:-1]):

            neighbouring_combinations = self._get_neighbouring_combinations(
                best_combination, step_size, previous_step_size
            )

            best_combination, best_score = self._find_best_parameters_iter(
                neighbouring_combinations
            )

        arguments = self._build_arguments(best_arm, best_combination)

        return arguments

    def _get_neighbouring_combinations(
        self,
        best_combination: Union[List[float16], ndarray],
        step_size: int,
        previous_step_size: int,
    ) -> ndarray:
        new_sets = []
        for real_value, range_ in zip(
            best_combination, self._real_hyperparameter_ranges
        ):
            real_value_index = np.argmax(range_ == real_value)

            start_index = real_value_index - previous_step_size
            start_index = max(start_index, 0)
            end_index = min(real_value_index + previous_step_size, len(range_))

            adjusted_step_size = min(int(len(range_) / 2), step_size)
            new_set = list(range_[start_index:end_index:adjusted_step_size])
            if np.max(new_set) < real_value:
                new_set.append(real_value)

            new_sets.append(new_set)
        neighbouring_combinations = np.array(list(itertools.product(*new_sets)))

        return neighbouring_combinations

    def save_plots(self, path_stem):
        """Create and save plots that summarise GEC trajectory

        Parameters
        ----------
            path_stem : str
                path to folder + file name root to save plots to
        """
        figs = self.plot_gec()
        self._write_figures(figs, path_stem)

    def _write_figures(self, figs, path_stem):
        for plot_name, fig in figs.items():
            fig.savefig(f"{path_stem}_{plot_name}.png")

    def plot_gec(self) -> Dict[str, Figure]:
        """Create figures to summarise GEC trajectory

        Returns
        -------
            figs : dict[str, fig]
                a dictionary of figures
        """

        figs = {}

        fig, axes = plt.subplots(
            nrows=2, ncols=2, sharex=True, sharey=False, figsize=(12, 12)
        )
        ax1, ax2, ax3, ax4 = axes.flatten()

        x = np.arange(len(self.hyperparameter_scores_["means"]))
        self._plot_mean_prediction_and_mean_variance(ax1, x)
        self._plot_prediction_std_and_variance_std(ax2, x)
        self._plot_prediction_mean_variance_correlation(ax3, x)
        self._plot_linear_scaled_parameter_samples(ax4)

        figs["parameters"] = fig

        return figs

    def _plot_mean_prediction_and_mean_variance(self, ax: Axes, x: ndarray) -> None:
        gp_mean_prediction = [np.mean(x) for x in self.hyperparameter_scores_["means"]]
        gp_mean_sigma = [np.mean(x) for x in self.hyperparameter_scores_["sigmas"]]

        ax.plot(x, gp_mean_prediction, label="mean_prediction")
        ax.plot(x, gp_mean_sigma, label="mean_sigma")
        ax.legend(loc="upper right")

    def _plot_prediction_std_and_variance_std(self, ax: Axes, x: ndarray) -> None:
        gp_prediction_variance = [
            np.std(x) for x in self.hyperparameter_scores_["means"]
        ]
        gp_sigma_variance = [np.std(x) for x in self.hyperparameter_scores_["sigmas"]]

        ax.plot(x, gp_prediction_variance, label="prediction_variance")
        ax.plot(x, gp_sigma_variance, label="sigma_variance")
        ax.legend(loc="lower right")

    def _plot_prediction_mean_variance_correlation(self, ax: Axes, x: ndarray) -> None:
        correlation = [
            np.corrcoef(
                self.hyperparameter_scores_["means"][i],
                self.hyperparameter_scores_["sigmas"][i],
            )[0, 1]
            for i in x
        ]

        ax.plot(
            x,
            correlation,
            label="corr(preds mean, preds variance)",
        )
        ax.legend(loc="lower right")

    def _plot_linear_scaled_parameter_samples(self, ax: Axes) -> None:
        inputs_ = np.array(self.hyperparameter_scores_["inputs"])
        x = np.arange(inputs_.shape[0])

        for i in range(inputs_.shape[1]):
            ax.plot(x, inputs_[:, i], label=self._real_hyperparameter_names[i])



import copy
import inspect
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
from lightgbm import LGBMRegressor
from lightgbm.basic import Booster, LightGBMError
from lightgbm.compat import SKLEARN_INSTALLED
from lightgbm.sklearn import (
    LGBMModel,
    _EvalResultDict,
    _LGBM_BoosterBestScoreType,
    _LGBM_CategoricalFeatureConfiguration,
    _LGBM_FeatureNameConfiguration,
    _LGBM_InitScoreType,
    _LGBM_LabelType,
    _LGBM_ScikitCustomObjectiveFunction,
    _LGBM_ScikitEvalMetricType,
    _LGBM_ScikitMatrixLike,
    _LGBM_ScikitValidSet,
    _LGBM_WeightType,
)
from numpy import float64

from .gec_base import GECBase


class LightGER(LGBMRegressor, GECBase):
    def __init__(
        self,
        boosting_type: str = "gbdt",
        num_leaves: int = 31,
        max_depth: int = -1,
        learning_rate: float = 0.1,
        n_estimators: int = 100,
        subsample_for_bin: int = 200000,
        objective: Optional[Union[str, _LGBM_ScikitCustomObjectiveFunction]] = None,
        class_weight: Optional[Union[Dict, str]] = None,
        min_split_gain: float = 0.0,
        min_child_weight: float = 1e-3,
        min_child_samples: int = 20,
        subsample: float = 1.0,
        subsample_freq: int = 0,
        colsample_bytree: float = 1.0,
        reg_alpha: float = 0.0,
        reg_lambda: float = 0.0,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        n_jobs: Optional[int] = None,
        importance_type: str = "split",
        frozen: bool = False,
        **kwargs,
    ) -> None:
        adapted_lgbm_params = (
            str(inspect.signature(LGBMRegressor.__init__))
            .replace(
                "importance_type: str = 'split'",
                "importance_type: str = 'split', frozen: bool = False",
            )
            .replace("**kwargs)", "**kwargs) -> None")
        )
        gec_params = str(inspect.signature(LightGER.__init__))
        assert (
            adapted_lgbm_params == gec_params
        ), f"{gec_params = } \n not equal to \n {adapted_lgbm_params = }"

        r"""Construct a gradient boosting model.

        Parameters
        ----------
        boosting_type : str, optional (default='gbdt')
            'gbdt', traditional Gradient Boosting Decision Tree.
            'dart', Dropouts meet Multiple Additive Regression Trees.
            'rf', Random Forest.
        num_leaves : int, optional (default=31)
            Maximum tree leaves for base learners.
        max_depth : int, optional (default=-1)
            Maximum tree depth for base learners, <=0 means no limit.
        learning_rate : float, optional (default=0.1)
            Boosting learning rate.
            You can use ``callbacks`` parameter of ``fit`` method to shrink/adapt learning rate
            in training using ``reset_parameter`` callback.
            Note, that this will ignore the ``learning_rate`` argument in training.
        n_estimators : int, optional (default=100)
            Number of boosted trees to fit.
        subsample_for_bin : int, optional (default=200000)
            Number of samples for constructing bins.
        objective : str, callable or None, optional (default=None)
            Specify the learning task and the corresponding learning objective or
            a custom objective function to be used (see note below).
            Default: 'regression' for LGBMRegressor, 'binary' or 'multiclass' for LGBMClassifier, 'lambdarank' for LGBMRanker.
        class_weight : dict, 'balanced' or None, optional (default=None)
            Weights associated with classes in the form ``{class_label: weight}``.
            Use this parameter only for multi-class classification task;
            for binary classification task you may use ``is_unbalance`` or ``scale_pos_weight`` parameters.
            Note, that the usage of all these parameters will result in poor estimates of the individual class probabilities.
            You may want to consider performing probability calibration
            (https://scikit-learn.org/stable/modules/calibration.html) of your model.
            The 'balanced' mode uses the values of y to automatically adjust weights
            inversely proportional to class frequencies in the input data as ``n_samples / (n_classes * np.bincount(y))``.
            If None, all classes are supposed to have weight one.
            Note, that these weights will be multiplied with ``sample_weight`` (passed through the ``fit`` method)
            if ``sample_weight`` is specified.
        min_split_gain : float, optional (default=0.)
            Minimum loss reduction required to make a further partition on a leaf node of the tree.
        min_child_weight : float, optional (default=1e-3)
            Minimum sum of instance weight (Hessian) needed in a child (leaf).
        min_child_samples : int, optional (default=20)
            Minimum number of data needed in a child (leaf).
        subsample : float, optional (default=1.)
            Subsample ratio of the training instance.
        subsample_freq : int, optional (default=0)
            Frequency of subsample, <=0 means no enable.
        colsample_bytree : float, optional (default=1.)
            Subsample ratio of columns when constructing each tree.
        reg_alpha : float, optional (default=0.)
            L1 regularization term on weights.
        reg_lambda : float, optional (default=0.)
            L2 regularization term on weights.
        random_state : int, RandomState object or None, optional (default=None)
            Random number seed.
            If int, this number is used to seed the C++ code.
            If RandomState object (numpy), a random integer is picked based on its state to seed the C++ code.
            If None, default seeds in C++ code are used.
        n_jobs : int or None, optional (default=None)
            Number of parallel threads to use for training (can be changed at prediction time by
            passing it as an extra keyword argument).

            For better performance, it is recommended to set this to the number of physical cores
            in the CPU.

            Negative integers are interpreted as following joblib's formula (n_cpus + 1 + n_jobs), just like
            scikit-learn (so e.g. -1 means using all threads). A value of zero corresponds the default number of
            threads configured for OpenMP in the system. A value of ``None`` (the default) corresponds
            to using the number of physical cores in the system (its correct detection requires
            either the ``joblib`` or the ``psutil`` util libraries to be installed).

            .. versionchanged:: 4.0.0

        importance_type : str, optional (default='split')
            The type of feature importance to be filled into ``feature_importances_``.
            If 'split', result contains numbers of times the feature is used in a model.
            If 'gain', result contains total gains of splits which use the feature.
        **kwargs
            Other parameters for the model.
            Check http://lightgbm.readthedocs.io/en/latest/Parameters.html for more parameters.

            .. warning::

                \*\*kwargs is not supported in sklearn, it may cause unexpected issues.

        Note
        ----
        A custom objective function can be provided for the ``objective`` parameter.
        In this case, it should have the signature
        ``objective(y_true, y_pred) -> grad, hess``,
        ``objective(y_true, y_pred, weight) -> grad, hess``
        or ``objective(y_true, y_pred, weight, group) -> grad, hess``:

            y_true : numpy 1-D array of shape = [n_samples]
                The target values.
            y_pred : numpy 1-D array of shape = [n_samples] or numpy 2-D array of shape = [n_samples, n_classes] (for multi-class task)
                The predicted values.
                Predicted values are returned before any transformation,
                e.g. they are raw margin instead of probability of positive class for binary task.
            weight : numpy 1-D array of shape = [n_samples]
                The weight of samples. Weights should be non-negative.
            group : numpy 1-D array
                Group/query data.
                Only used in the learning-to-rank task.
                sum(group) = n_samples.
                For example, if you have a 100-document dataset with ``group = [10, 20, 40, 10, 10, 10]``, that means that you have 6 groups,
                where the first 10 records are in the first group, records 11-30 are in the second group, records 31-70 are in the third group, etc.
            grad : numpy 1-D array of shape = [n_samples] or numpy 2-D array of shape = [n_samples, n_classes] (for multi-class task)
                The value of the first order derivative (gradient) of the loss
                with respect to the elements of y_pred for each sample point.
            hess : numpy 1-D array of shape = [n_samples] or numpy 2-D array of shape = [n_samples, n_classes] (for multi-class task)
                The value of the second order derivative (Hessian) of the loss
                with respect to the elements of y_pred for each sample point.

        For multi-class task, y_pred is a numpy 2-D array of shape = [n_samples, n_classes],
        and grad and hess should be returned in the same format.
        """

        if not SKLEARN_INSTALLED:
            raise LightGBMError(
                "scikit-learn is required for lightgbm.sklearn. "
                "You must install scikit-learn and restart your session to use this module."
            )

        self.boosting_type = boosting_type
        self.objective = objective
        self.num_leaves = num_leaves
        self.max_depth = max_depth
        self.learning_rate = learning_rate
        self.n_estimators = n_estimators
        self.subsample_for_bin = subsample_for_bin
        self.min_split_gain = min_split_gain
        self.min_child_weight = min_child_weight
        self.min_child_samples = min_child_samples
        self.subsample = subsample
        self.subsample_freq = subsample_freq
        self.colsample_bytree = colsample_bytree
        self.reg_alpha = reg_alpha
        self.reg_lambda = reg_lambda
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.importance_type = importance_type
        self._Booster: Optional[Booster] = None
        self._evals_result: _EvalResultDict = {}
        self._best_score: _LGBM_BoosterBestScoreType = {}
        self._best_iteration: int = -1
        self._other_params: Dict[str, Any] = {}
        self._objective = objective
        self.class_weight = class_weight
        self._class_weight: Optional[Union[Dict, str]] = None
        self._class_map: Optional[Dict[int, int]] = None
        self._n_features: int = -1
        self._n_features_in: int = -1
        self._classes: Optional[np.ndarray] = None
        self._n_classes: int = -1
        self.set_params(**kwargs)
        non_optimized_init_args = [
            "max_depth",
            "subsample_for_bin",
            "objective",
            "class_weight",
            "min_split_gain",
            "random_state",
            "n_jobs",
            "importance_type",
        ]
        optimization_candidate_init_args = [
            "learning_rate",
            "n_estimators",
            "num_leaves",
            "reg_alpha",
            "reg_lambda",
            "min_child_samples",
            "min_child_weight",
            "colsample_bytree",  # feature_fraction
            "subsample",
        ]
        categorical_hyperparameters = [
            ("boosting_type", ["gbdt", "dart", "rf"]),
        ]
        self._gec_init(
            frozen,
            non_optimized_init_args,
            optimization_candidate_init_args,
            categorical_hyperparameters,
        )

    def fit(
        self,
        X: _LGBM_ScikitMatrixLike,
        y: _LGBM_LabelType,
        n_iter: int = 50,
        fixed_hyperparameters: List[str] = ["n_estimators"],
        sample_weight: Optional[_LGBM_WeightType] = None,
        init_score: Optional[_LGBM_InitScoreType] = None,
        eval_set: Optional[List[_LGBM_ScikitValidSet]] = None,
        eval_names: Optional[List[str]] = None,
        eval_sample_weight: Optional[List[_LGBM_WeightType]] = None,
        eval_init_score: Optional[List[_LGBM_InitScoreType]] = None,
        eval_metric: Optional[_LGBM_ScikitEvalMetricType] = None,
        feature_name: _LGBM_FeatureNameConfiguration = "auto",
        categorical_feature: _LGBM_CategoricalFeatureConfiguration = "auto",
        callbacks: Optional[List[Callable]] = None,
        init_model: Optional[Union[str, Path, Booster, LGBMModel]] = None,
    ) -> "LightGER":
        """Docstring is inherited from the LGBMRegressor.

        Except for

        Parameters:
        ----------
            n_iter : int
                number of optimization steps
            fixed_hyperparameters : list[str]
                list of hyperparameters that should not be optimised
        """

        self.gec_fit_params_ = {
            "sample_weight": sample_weight,
            "init_score": init_score,
            "eval_set": eval_set,
            "eval_names": eval_names,
            "eval_sample_weight": eval_sample_weight,
            "eval_init_score": eval_init_score,
            "eval_metric": eval_metric,
            "feature_name": feature_name,
            "categorical_feature": categorical_feature,
            "callbacks": callbacks,
            "init_model": init_model,
        }
        self._fit_inner(X, y, n_iter, fixed_hyperparameters)

    def set_params(self, **kwargs) -> None:
        if "frozen" in kwargs:
            setattr(self, "frozen", kwargs.pop("frozen"))
        super().set_params(**kwargs)

    def get_params(
        self, deep: bool = True
    ) -> Dict[str, Optional[Union[str, float, int, bool]]]:
        if hasattr(self, "best_params_") and self.best_params_ is not None:
            params = copy.deepcopy(self.best_params_)
        else:
            params = super().get_params(deep)
        params["frozen"] = self.frozen

        return params

    def _fit_best_params(self, X: _LGBM_ScikitMatrixLike, y: _LGBM_LabelType) -> None:

        if hasattr(self, "best_params") and self.best_params_ is not None:
            for k, v in self.best_params_.items():
                setattr(self, k, v)
            setattr(self, "random_state", 101)

        super().fit(X, y, **self.gec_fit_params_)

    def score_single_iteration(
        self,
        X: _LGBM_ScikitMatrixLike,
        y: _LGBM_LabelType,
        params: Dict[str, Optional[Union[str, float, int, float64]]],
    ) -> float64:
        return self._calculate_cv_score(X, y, params, LGBMRegressor)

    def retrieve_hyperparameter(
        self, hyperparameter: str
    ) -> Optional[Union[int, str, float]]:
        return getattr(self, hyperparameter)

    def _process_arguments(
        self, arguments: Dict[str, Optional[Union[int, float, str]]]
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        args = copy.deepcopy(arguments)
        bagging = args["gec_bagging"] == "gec_bagging_yes"
        if bagging:
            assert "subsample" in args
            args["subsample_freq"] = 1
        else:
            del args["subsample"]
        del args["gec_bagging"]
        return args



import copy
import inspect
import sys
from typing import Callable, Dict, List, Optional, Union

import numpy as np
from catboost import CatBoostClassifier
from numpy import float64, ndarray
from six import integer_types, iteritems, string_types

from .gec_base import GECBase


class CatGEC(CatBoostClassifier, GECBase):
    def __init__(
        self,
        iterations=None,
        learning_rate=None,
        depth=None,
        l2_leaf_reg=None,
        model_size_reg=None,
        rsm=None,
        loss_function=None,
        border_count=None,
        feature_border_type=None,
        per_float_feature_quantization=None,
        input_borders=None,
        output_borders=None,
        fold_permutation_block=None,
        od_pval=None,
        od_wait=None,
        od_type=None,
        nan_mode=None,
        counter_calc_method=None,
        leaf_estimation_iterations=None,
        leaf_estimation_method=None,
        thread_count=None,
        random_seed=None,
        use_best_model=None,
        best_model_min_trees=None,
        verbose=None,
        silent=None,
        logging_level=None,
        metric_period=None,
        ctr_leaf_count_limit=None,
        store_all_simple_ctr=None,
        max_ctr_complexity=None,
        has_time=None,
        allow_const_label=None,
        target_border=None,
        classes_count=None,
        class_weights=None,
        auto_class_weights=None,
        class_names=None,
        one_hot_max_size=None,
        random_strength=None,
        name=None,
        ignored_features=None,
        train_dir=None,
        custom_loss=None,
        custom_metric=None,
        eval_metric=None,
        bagging_temperature=None,
        save_snapshot=None,
        snapshot_file=None,
        snapshot_interval=None,
        fold_len_multiplier=None,
        used_ram_limit=None,
        gpu_ram_part=None,
        pinned_memory_size=None,
        allow_writing_files=None,
        final_ctr_computation_mode=None,
        approx_on_full_history=None,
        boosting_type=None,
        simple_ctr=None,
        combinations_ctr=None,
        per_feature_ctr=None,
        ctr_description=None,
        ctr_target_border_count=None,
        task_type=None,
        device_config=None,
        devices=None,
        bootstrap_type=None,
        subsample=None,
        mvs_reg=None,
        sampling_unit=None,
        sampling_frequency=None,
        dev_score_calc_obj_block_size=None,
        dev_efb_max_buckets=None,
        sparse_features_conflict_fraction=None,
        max_depth=None,
        n_estimators=None,
        num_boost_round=None,
        num_trees=None,
        colsample_bylevel=None,
        random_state=None,
        reg_lambda=None,
        objective=None,
        eta=None,
        max_bin=None,
        scale_pos_weight=None,
        gpu_cat_features_storage=None,
        data_partition=None,
        metadata=None,
        early_stopping_rounds=None,
        cat_features=None,
        grow_policy=None,
        min_data_in_leaf=None,
        min_child_samples=None,
        max_leaves=None,
        num_leaves=None,
        score_function=None,
        leaf_estimation_backtracking=None,
        ctr_history_unit=None,
        monotone_constraints=None,
        feature_weights=None,
        penalties_coefficient=None,
        first_feature_use_penalties=None,
        per_object_feature_penalties=None,
        model_shrink_rate=None,
        model_shrink_mode=None,
        langevin=None,
        diffusion_temperature=None,
        posterior_sampling=None,
        boost_from_average=None,
        text_features=None,
        tokenizers=None,
        dictionaries=None,
        feature_calcers=None,
        text_processing=None,
        embedding_features=None,
        callback=None,
        eval_fraction=None,
        fixed_binary_splits=None,
        frozen=False,
    ):
        adapted_cat_params = str(
            inspect.signature(CatBoostClassifier.__init__)
        ).replace(
            "fixed_binary_splits=None)",
            "fixed_binary_splits=None, frozen=False)",
        )
        catgec_params = str(inspect.signature(CatGEC.__init__))
        assert (
            adapted_cat_params == catgec_params
        ), f"{catgec_params = } \n not equal to \n {adapted_cat_params = }"

        params = {}
        not_params = [
            "not_params",
            "self",
            "params",
            "__class__",
            "catgec_params",
            "adapted_cat_params",
            "frozen",
        ]
        for key, value in iteritems(locals().copy()):
            if key not in not_params and value is not None:
                params[key] = value

        super(CatBoostClassifier, self).__init__(params)

        non_optimized_init_args = [
            "depth",
            "model_size_reg",
            "loss_function",
            "border_count",
            "feature_border_type",
            "per_float_feature_quantization",
            "input_borders",
            "output_borders",
            "fold_permutation_block",
            "od_pval",
            "od_wait",
            "od_type",
            "nan_mode",
            "counter_calc_method",
            "leaf_estimation_iterations",
            "leaf_estimation_method",
            "thread_count",
            "random_seed",
            "use_best_model",
            "best_model_min_trees",
            "verbose",
            "silent",
            "logging_level",
            "metric_period",
            "ctr_leaf_count_limit",
            "store_all_simple_ctr",
            "max_ctr_complexity",
            "has_time",
            "allow_const_label",
            "target_border",
            "one_hot_max_size",
            "random_strength",
            "name",
            "ignored_features",
            "train_dir",
            "custom_metric",
            "eval_metric",
            "bagging_temperature",
            "save_snapshot",
            "snapshot_file",
            "snapshot_interval",
            "fold_len_multiplier",
            "used_ram_limit",
            "gpu_ram_part",
            "pinned_memory_size",
            "allow_writing_files",
            "final_ctr_computation_mode",
            "approx_on_full_history",
            "simple_ctr",
            "combinations_ctr",
            "per_feature_ctr",
            "ctr_description",
            "ctr_target_border_count",
            "task_type",
            "device_config",
            "devices",
            "mvs_reg",
            "sampling_frequency",
            "sampling_unit",
            "dev_score_calc_obj_block_size",
            "dev_efb_max_buckets",
            "sparse_features_conflict_fraction",
            "max_depth",
            "random_state",
            "objective",
            "max_bin",
            "gpu_cat_features_storage",
            "data_partition",
            "metadata",
            "early_stopping_rounds",
            "cat_features",
            "grow_policy",
            "score_function",
            "leaf_estimation_backtracking",
            "ctr_history_unit",
            "monotone_constraints",
            "feature_weights",
            "penalties_coefficient",
            "first_feature_use_penalties",
            "per_object_feature_penalties",
            "model_shrink_rate",
            "model_shrink_mode",
            "langevin",
            "diffusion_temperature",
            "posterior_sampling",
            "boost_from_average",
            "text_features",
            "tokenizers",
            "dictionaries",
            "feature_calcers",
            "text_processing",
            "embedding_features",
            "eval_fraction",
            "fixed_binary_splits",
            "num_leaves",
        ]
        optimization_candidate_init_args = [
            "learning_rate",
            "n_estimators",
            "reg_lambda",
            "min_child_samples",
            "colsample_bylevel",  # feature_fraction
            "subsample",
        ]
        categorical_hyperparameters = [
            ("boosting_type", ["Plain"]),
            ("bootstrap_type", ["Bayesian", "Bernoulli", "MVS", "No"]),
        ]
        self._gec_init(
            frozen,
            non_optimized_init_args,
            optimization_candidate_init_args,
            categorical_hyperparameters,
        )

    def fit(
        self,
        X,
        y=None,
        n_iter=50,
        fixed_hyperparameters=["n_estimators"],
        cat_features=None,
        text_features=None,
        embedding_features=None,
        sample_weight=None,
        baseline=None,
        use_best_model=None,
        eval_set=None,
        verbose=None,
        logging_level=None,
        plot=False,
        plot_file=None,
        column_description=None,
        verbose_eval=None,
        metric_period=None,
        silent=None,
        early_stopping_rounds=None,
        save_snapshot=None,
        snapshot_file=None,
        snapshot_interval=None,
        init_model=None,
        callbacks=None,
        log_cout=sys.stdout,
        log_cerr=sys.stderr,
    ):
        self.gec_fit_params_ = {
            "cat_features": cat_features,
            "text_features": text_features,
            "embedding_features": embedding_features,
            "sample_weight": sample_weight,
            "baseline": baseline,
            "use_best_model": use_best_model,
            "eval_set": eval_set,
            "verbose": verbose,
            "logging_level": logging_level,
            "plot": plot,
            "plot_file": plot_file,
            "column_description": column_description,
            "verbose_eval": verbose_eval,
            "metric_period": metric_period,
            "silent": silent,
            "early_stopping_rounds": early_stopping_rounds,
            "save_snapshot": save_snapshot,
            "snapshot_file": snapshot_file,
            "snapshot_interval": snapshot_interval,
            "init_model": init_model,
            "callbacks": callbacks,
            "log_cout": log_cout,
            "log_cerr": log_cerr,
        }
        self._fit_inner(X, y, n_iter, fixed_hyperparameters)

    def set_params(self, **kwargs) -> None:
        if "frozen" in kwargs:
            setattr(self, "frozen", kwargs.pop("frozen"))
        super().set_params(**kwargs)

    def get_params(
        self, deep: bool = True
    ) -> Dict[str, Optional[Union[str, float, int, bool]]]:
        if hasattr(self, "best_params_") and self.best_params_ is not None:
            params = copy.deepcopy(self.best_params_)
        else:
            params = super().get_params(deep)
        params["frozen"] = self.frozen

        return {k: v for k, v in params.items() if v is not None}

    def _fit_best_params(self, X: ndarray, y: ndarray) -> None:

        if self.best_params_ is not None:
            for k, v in self.best_params_.items():
                if v is not None:
                    self._init_params[k] = v
            self._init_params["random_state"] = 101
            self._init_params["silent"] = True

        super().fit(X, y, **self.gec_fit_params_)

    def score_single_iteration(
        self,
        X: ndarray,
        y: ndarray,
        params: Dict[str, Optional[Union[str, float, int, float64]]],
    ) -> Union[float, float64]:

        if "subsample_freq" in params:
            del params["subsample_freq"]

        params["silent"] = True

        return self._calculate_cv_score(X, y, params, CatBoostClassifier)

    def retrieve_hyperparameter(self, hyperparameter: str) -> None:
        return self._init_params.get(hyperparameter, None)

    def _process_arguments(
        self, arguments: Dict[str, Optional[Union[int, float, str]]]
    ) -> Dict[str, Optional[Union[int, float, str]]]:
        args = copy.deepcopy(arguments)
        bagging = args["gec_bagging"] == "gec_bagging_yes"
        if bagging:
            assert "subsample" in args
        else:
            del args["subsample"]
        del args["gec_bagging"]
        return args



import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.ticker import LinearLocator
from sklearn.utils.extmath import cartesian


def visualise_1D_gaussian_process(gaussian):
    x = np.atleast_2d(np.linspace(0, 10, 100)).T
    y_pred, sigma = gaussian.predict(x, return_std=True)
    plt.figure(figsize=(15, 15))
    plt.plot(x, y_pred, "b-", label="Prediction")

    plt.fill(
        np.concatenate([x, x[::-1]]),
        np.concatenate([y_pred - 1.9600 * sigma, (y_pred + 1.9600 * sigma)[::-1]]),
        alpha=0.5,
        fc="b",
        ec="None",
        label="95% confidence interval",
    )
    plt.show()


def visualise_2D_gaussian_process(
    gaussian,
    X_range=np.arange(0, 10, 0.1),
    Y_range=np.arange(0, 10, 0.1),
    Z_range=np.arange(-3, 3, 1),
    plot_bounds=True,
):
    fig, ax = plt.subplots(subplot_kw={"projection": "3d"}, figsize=(15, 15))

    # Make data.
    X = np.arange(
        np.min(X_range), np.max(X_range), (np.max(X_range) - np.min(X_range)) / 100
    )
    Y = np.arange(
        np.min(Y_range), np.max(Y_range), (np.max(Y_range) - np.min(Y_range)) / 100
    )
    Z = np.arange(
        np.min(Z_range), np.max(Z_range), (np.max(Z_range) - np.min(Z_range)) / 100
    )
    R = cartesian(np.array([X, Y]))
    X, Y = np.meshgrid(X, Y)
    Z, sigma = gaussian.predict(R, return_std=True)
    Z = Z.reshape((100, 100))
    sigma = sigma.reshape((100, 100))

    # Plot the surface.
    surf = ax.plot_surface(X, Y, Z, cmap=cm.coolwarm, linewidth=0, antialiased=False)
    if plot_bounds:
        surf2 = ax.plot_surface(
            X, Y, Z + 1.9600 * sigma, alpha=0.2, linewidth=0, antialiased=False
        )
        surf3 = ax.plot_surface(
            X, Y, Z - 1.9600 * sigma, alpha=0.2, linewidth=0, antialiased=False
        )
    # Customize the z axis.
    ax.set_zlim(np.min(Z_range), np.max(Z_range))
    ax.zaxis.set_major_locator(LinearLocator(10))
    # A StrMethodFormatter is used automatically
    ax.zaxis.set_major_formatter("{x:.02f}")
    # Add a color bar which maps values to colors.
    fig.colorbar(surf, shrink=0.5, aspect=5)

    plt.show()
